// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//--------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//--------------------------------------------------------------------------
//
//	Evaluator.cc --- implementation of Evaluator class
//
//	DESCRIPTION:
//		Defines of functions bound to the operators recognized
//		by the expression scanner. For a list of operators/functions,
//		see Selector.h
//
//	CHANGE HISTORY:
//		UL, 1997-12-17: 1.0.0: - first version
//		UL, 1998-04-09: 1.1.0: - ifthenelse added
//		UL, 1998-05-19: 1.2.0: - overhaul; can now instantiate several
//								 Evaluator objects within the same scope;
//								 for this global lex/yacc variables are saved
//								 in constructor and restored in destructor
//		UL, 1998-10-07: 1.3.0: - some modifcations to support WCS in mask
//								 filter
//		UL, 1999-10-14: 1.4.0: - added implementation of isValidExpression()
//		UL, 2000-07-03: 1.5.0: - several changes related to support for
//								 string columns
//		UL, 2002-12-17: 1.6.0: - fixed bug in Evaluator destructor
//
//	AUTHOR:
//		U. Lammers (UL), Dec 17 11:48:12 MET DST 1997
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/selcore/Evaluator.cc,v 1.4 2009/06/26 15:51:59 aibarra Exp $

#include "basicdefs.h"
#include "Selector.h"
#include "SelectError.h"
#include "Evaluator.h"

using namespace std;
using namespace SelectError;

//
//	----------------------( lexer/parser interface )-------------------------
//
namespace SelectParser {
	void lexinit(const string &expression);	//	initialize the lexer
	int	yyparse();							//	parse the expression
}


bool Evaluator::isValidExpression(const string &expression)
{
	bool res = true;

	selectErrorHandler.exceptions(true);	//	enable exceptions
	try {
		Evaluator ev(expression);
	} catch (SelectException) {
		res = false;
	}

	selectErrorHandler.oldHandling();

	return res;
} // Evaluator::isValidExpression


Evaluator::Evaluator(const string &expression,
					 const TableDescriptor *tableDescriptor)
{
	//	save current symbol table and expression tree to make multiple
	//	Evaluator instantiations possible
	//	they are restored in the destructor
//	_prevSelectorGlobals	= selectorGlobals;
	_flagColumnNameSet		= false;
	_flagColumnName			= SELECT_COLUMN;
	
	_selectorGlobals.evaluator = this;
	_selectorGlobals.symbolTable = new SymbolTable;	//	create new symbol table
	_selectorGlobals.tableDescriptor = tableDescriptor;
	selectorGlobals = &_selectorGlobals;

	std::string val = "true";
	
	if (expression.length() != 0)
	  val = expression;

	SelectParser::lexinit(val);
	if (SelectParser::yyparse()) {
		selerrstr<<"Syntax error in expression: `"<<expression<<"'"
				 <<selerror("SyntaxError", 1);
		SELTHROW(InvalidExpression);
	}
} // Evaluator::Evaluator()


NumericSymbols Evaluator::numericIdentifiers() const
{
	NumericSymbols	symbols;

	unsigned symTabLen = _selectorGlobals.symbolTable->size();
	for (unsigned n=0; n<symTabLen; ++n) {
		Identifier *id = (*_selectorGlobals.symbolTable)[n];
		string idName(id->name());
		if (id->isNumeric() &&
			idName != ROWNUMBER_ID &&		//	handled in rowNumber()
			idName != SELECT_MASK &&		//	handled in setFlagData()
			(idName != _flagColumnName || _flagColumnNameSet)) {
				//	that's a true identifier
				symbols.name.push_back(idName);
				symbols.valuePtr.push_back(&id->_value);
		}
	}

 	return symbols;
} // Evaluator::numericIdentifiers()


StringSymbols Evaluator::stringIdentifiers() const
{
	StringSymbols	symbols;

	unsigned symTabLen = _selectorGlobals.symbolTable->size();
	for (unsigned n=0; n<symTabLen; ++n) {
		Identifier *id = (*_selectorGlobals.symbolTable)[n];
		if (!id->isNumeric()) {
			symbols.name.push_back(id->name());
			symbols.maxLength.push_back(id->_textLength);
			symbols.textPtr.push_back(id->_text);
		}
	}

 	return symbols;
} // Evaluator::stringIdentifiers()


double *Evaluator::rowNumber() const
{
	unsigned symTabLen = _selectorGlobals.symbolTable->size();
	for (unsigned n=0; n<symTabLen; ++n) {
		Identifier *id = (*_selectorGlobals.symbolTable)[n];
		if (id->name() == ROWNUMBER_ID)
			return &id->_value;
	}

	return 0;
}


bool Evaluator::isLogic() const
{
	return _selectorGlobals.isSelExprLogic;
} // Evaluator::isSelExprLogic()


double Evaluator::evalArith() const
{
	if (!_selectorGlobals.isSelExprLogic) {
		selectorGlobals = &_selectorGlobals;
		return _selectorGlobals.selExpression->arith(
											_selectorGlobals.selExpression);
		//	note: we do not need to test for selExpression!=0
		//	because a valid expression must have been created in the
		//	constructor
	} else {
		selerrstr<<"Expression cannot be evaluated in arithmetic context"
				 <<selerror("NoArithmeticContext", 2);
		SELTHROW(NoArithContext);
	}

	return 0.;
} // Evaluator::evalArith()


bool Evaluator::evalLogic() const
{
	if (_selectorGlobals.isSelExprLogic) {
		selectorGlobals = &_selectorGlobals;
		return _selectorGlobals.selExpression->logic(
											_selectorGlobals.selExpression);
		//	note: we do not need to test for selExpression!=0
		//	because a valid expression must have been created in the
		//	constructor
	} else {
		selerrstr<<"Expression cannot be evaluated in logical context"
				 <<selerror("NoLogicalContext", 2);
		SELTHROW(NoLogicContext);
	}

	return false;
} // Evaluator::evalLogic()


void Evaluator::setFlagData(const string &flagColumnName,
							unsigned short flagBit) const
{
	//
	//	here we need to find the symbols SELECT_COLUMN and SELECT_MASK
	//	in the symbol table and replace them with the right values
	//
	unsigned symTabLen = _selectorGlobals.symbolTable->size();
	for (unsigned n=0; n<symTabLen; ++n) {
		Identifier *id = (*_selectorGlobals.symbolTable)[n];
		string idName(id->name());
		if (idName == SELECT_COLUMN && flagColumnName.length()) {
			//	found the flag column - rename the symbol
			(*_selectorGlobals.symbolTable)[n]->name(flagColumnName);
			_flagColumnNameSet = true;
			_flagColumnName = flagColumnName;
		} else if (idName == SELECT_MASK)
		{
			if (flagColumnName.length())
				//	asign the right value
				id->_value = DBL(1<<flagBit);
			else
				id->_value = 0.;
		}
	}
} // Evaluator::setFlagData()


Evaluator::~Evaluator()
{
	if (!_selectorGlobals.selExpression->isIdentifierNode())
		//	id's get deleted as part of SymbolTable cleanup
		delete _selectorGlobals.selExpression;
	delete _selectorGlobals.symbolTable;

	selectorGlobals = &selectorGlobalsInstance;
}
